from __future__ import annotations

import functools
import json
import os
import signal
import threading
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from sys import argv
from time import sleep
from typing import Any
from unittest.mock import ANY, MagicMock, patch

import pytest
from blueapi.core import BlueskyContext
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.baton import Baton
from dodal.devices.zebra.zebra import Zebra
from flask.testing import FlaskClient
from ophyd_async.testing import set_mock_value

from mx_bluesky.common.external_interaction.alerting.log_based_service import (
    LoggingAlertService,
)
from mx_bluesky.common.utils.context import (
    device_composite_from_context,
    find_device_in_context,
)
from mx_bluesky.common.utils.exceptions import WarningException
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.__main__ import (
    Actions,
    RunExperiment,
    Status,
    create_app,
    initialise_globals,
    main,
    setup_context,
)
from mx_bluesky.hyperion.baton_handler import HYPERION_USER
from mx_bluesky.hyperion.experiment_plans.experiment_registry import PLAN_REGISTRY
from mx_bluesky.hyperion.parameters.cli import (
    HyperionArgs,
    HyperionMode,
    parse_cli_args,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan
from mx_bluesky.hyperion.plan_runner import PlanRunner
from mx_bluesky.hyperion.runner import GDARunner

from ...conftest import mock_beamline_module_filepaths, raw_params_from_file
from .conftest import AGAMEMNON_WAIT_INSTRUCTION

FGS_ENDPOINT = "/pin_tip_centre_then_xray_centre/"
START_ENDPOINT = FGS_ENDPOINT + Actions.START.value
STOP_ENDPOINT = Actions.STOP.value
STATUS_ENDPOINT = Actions.STATUS.value
SHUTDOWN_ENDPOINT = Actions.SHUTDOWN.value
FLUSH_LOGS_ENDPOINT = "flush_debug_log"
TEST_BAD_PARAM_ENDPOINT = "/fgs_real_params/" + Actions.START.value

SECS_PER_RUNENGINE_LOOP = 0.1
RUNENGINE_TAKES_TIME_TIMEOUT = 15

"""
Every test in this file which uses the test_env fixture should either:
    - set RE_takes_time to false
    or
    - set an error on the mock run engine
In order to avoid threads which get left alive forever after test completion
"""


autospec_patch = functools.partial(patch, autospec=True, spec_set=True)
_MULTILINE_MESSAGE = "This is a\nmultiline log\nmessage."


@pytest.fixture()
def test_params(tmp_path):
    return json.dumps(
        raw_params_from_file(
            "tests/test_data/parameter_json_files/good_test_pin_centre_then_xray_centre_parameters.json",
            tmp_path,
        )
    )


class MockRunEngine:
    def __init__(self, test_name):
        self.RE_takes_time = True
        self.aborting_takes_time = False
        self.error: Exception | None = None
        self.test_name = test_name

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        time = 0.0
        while self.RE_takes_time:
            if self.error:
                raise self.error
            if time > RUNENGINE_TAKES_TIME_TIMEOUT:
                raise TimeoutError(
                    f'Mock RunEngine thread for test "{self.test_name}" spun too long'
                    "without an error. Most likely you should initialise with "
                    "RE_takes_time=false, or set RE.error from another thread."
                )
            sleep(SECS_PER_RUNENGINE_LOOP)
            time += SECS_PER_RUNENGINE_LOOP
        if self.error:
            raise self.error

    def abort(self):
        while self.aborting_takes_time:
            if self.error:
                raise self.error
            sleep(SECS_PER_RUNENGINE_LOOP)
        self.RE_takes_time = False

    def subscribe(self, *args):
        pass

    def unsubscribe(self, *args):
        pass


@dataclass
class ClientAndRunEngine:
    client: FlaskClient
    mock_run_engine: MockRunEngine


def mock_dict_values(d: dict):
    return {k: MagicMock() if k == "setup" or k == "run" else v for k, v in d.items()}


TEST_EXPTS = {
    "test_experiment": {
        "setup": MagicMock(),
        "param_type": MagicMock(),
    },
    "fgs_real_params": {
        "setup": MagicMock(),
        "param_type": HyperionSpecifiedThreeDGridScan,
    },
}


@pytest.fixture(autouse=True)
def mock_create_udc_server():
    with patch("mx_bluesky.hyperion.__main__.create_server_for_udc") as mock_udc_server:
        yield mock_udc_server


@pytest.fixture
def mock_setup_context(request: pytest.FixtureRequest):
    with (
        patch("mx_bluesky.hyperion.__main__.setup_context") as mock_setup_context,
        patch("mx_bluesky.hyperion.__main__.run_forever"),
    ):
        yield mock_setup_context


@pytest.fixture
def test_env(request: pytest.FixtureRequest, use_beamline_t01):
    mock_run_engine = MockRunEngine(test_name=repr(request))
    mock_context = BlueskyContext(run_engine=mock_run_engine)  # type: ignore
    real_plans_and_test_exps = dict(
        {k: mock_dict_values(v) for k, v in PLAN_REGISTRY.items()},  # type: ignore
        **TEST_EXPTS,  # type: ignore
    )
    mock_context.plan_functions = {  # type: ignore
        k: MagicMock() for k in real_plans_and_test_exps.keys()
    }

    with (
        patch.dict(
            "mx_bluesky.hyperion.__main__.PLAN_REGISTRY",
            real_plans_and_test_exps,
        ),
        patch(
            "mx_bluesky.hyperion.__main__.setup_context",
            MagicMock(return_value=mock_context),
        ),
    ):
        runner = GDARunner(mock_context)
        app = create_app(runner, {"TESTING": True})  # type: ignore

    runner_thread = threading.Thread(target=runner.wait_on_queue)
    runner_thread.start()
    with (
        app.test_client() as client,
        patch.dict(
            "mx_bluesky.hyperion.__main__.PLAN_REGISTRY",
            real_plans_and_test_exps,
        ),
    ):
        yield ClientAndRunEngine(client, mock_run_engine)

    runner.shutdown()
    runner_thread.join(timeout=3)
    del mock_run_engine


@pytest.fixture
def mock_flask_thread():
    # Prevent blocking when we attempt to join()
    with patch("mx_bluesky.hyperion.__main__.threading") as mock_threading:
        yield mock_threading


def wait_for_run_engine_status(
    client: FlaskClient,
    status_check: Callable[[str], bool] = lambda status: status != Status.BUSY.value,
    attempts=10,
):
    while attempts != 0:
        response = client.get(STATUS_ENDPOINT)
        response_json = json.loads(response.data)
        LOGGER.debug(
            f"Checking client status - response: {response_json}, attempts left={attempts}"
        )
        if status_check(response_json["status"]):
            return response_json
        else:
            attempts -= 1
            sleep(0.2)
    raise AssertionError("Run engine still busy")


def check_status_in_response(response_object, expected_result: Status):
    response_json = json.loads(response_object.data)
    assert response_json["status"] == expected_result.value, (
        f"{response_json['status']} != {expected_result.value}: {response_json.get('message')}"
    )


@pytest.mark.timeout(5)
def test_start_gives_success(test_env: ClientAndRunEngine, test_params):
    response = test_env.client.put(START_ENDPOINT, data=test_params)
    check_status_in_response(response, Status.SUCCESS)


@pytest.mark.timeout(4)
def test_getting_status_return_idle(test_env: ClientAndRunEngine, test_params):
    test_env.client.put(START_ENDPOINT, data=test_params)
    test_env.client.put(STOP_ENDPOINT)
    response = test_env.client.get(STATUS_ENDPOINT)
    check_status_in_response(response, Status.IDLE)


@pytest.mark.timeout(5)
def test_getting_status_after_start_sent_returns_busy(
    test_env: ClientAndRunEngine, test_params
):
    test_env.client.put(START_ENDPOINT, data=test_params)
    response = test_env.client.get(STATUS_ENDPOINT)
    check_status_in_response(response, Status.BUSY)


def test_putting_bad_plan_fails(test_env: ClientAndRunEngine, test_params):
    response = test_env.client.put("/bad_plan/start", data=test_params).json
    assert isinstance(response, dict)
    assert response.get("status") == Status.FAILED.value
    assert (
        response.get("message")
        == "PlanNotFound(\"Experiment plan 'bad_plan' not found in registry.\")"
    )
    test_env.mock_run_engine.abort()


def test_plan_with_no_params_fails(test_env: ClientAndRunEngine, test_params):
    response = test_env.client.put(
        "/test_experiment_no_internal_param_type/start", data=test_params
    ).json
    assert isinstance(response, dict)
    assert response.get("status") == Status.FAILED.value
    assert isinstance(message := response.get("message"), str)
    assert "'test_experiment_no_internal_param_type' not found in registry." in message
    test_env.mock_run_engine.abort()


@pytest.mark.timeout(7)
def test_sending_start_twice_fails(test_env: ClientAndRunEngine, test_params):
    test_env.client.put(START_ENDPOINT, data=test_params)
    response = test_env.client.put(START_ENDPOINT, data=test_params)
    check_status_in_response(response, Status.FAILED)


@pytest.mark.timeout(5)
def test_given_started_when_stopped_then_success_and_idle_status(
    test_env: ClientAndRunEngine, test_params
):
    test_env.mock_run_engine.aborting_takes_time = True
    test_env.client.put(START_ENDPOINT, data=test_params)
    response = test_env.client.put(STOP_ENDPOINT)
    check_status_in_response(response, Status.ABORTING)
    response = test_env.client.get(STATUS_ENDPOINT)
    check_status_in_response(response, Status.ABORTING)
    test_env.mock_run_engine.aborting_takes_time = False
    wait_for_run_engine_status(
        test_env.client, lambda status: status != Status.ABORTING
    )
    check_status_in_response(response, Status.ABORTING)


@pytest.mark.timeout(10)
def test_given_started_when_stopped_and_started_again_then_runs(
    test_env: ClientAndRunEngine, test_params
):
    test_env.client.put(START_ENDPOINT, data=test_params)
    test_env.client.put(STOP_ENDPOINT)
    test_env.mock_run_engine.RE_takes_time = True
    response = test_env.client.put(START_ENDPOINT, data=test_params)
    check_status_in_response(response, Status.SUCCESS)
    response = test_env.client.get(STATUS_ENDPOINT)
    check_status_in_response(response, Status.BUSY)
    test_env.mock_run_engine.RE_takes_time = False


@pytest.mark.timeout(5)
def test_when_started_n_returnstatus_interrupted_bc_RE_aborted_thn_error_reptd(
    test_env: ClientAndRunEngine, test_params
):
    test_env.mock_run_engine.aborting_takes_time = True
    test_env.client.put(START_ENDPOINT, data=test_params)
    test_env.client.put(STOP_ENDPOINT)
    test_env.mock_run_engine.error = Exception("D'Oh")
    response_json = wait_for_run_engine_status(
        test_env.client, lambda status: status != Status.ABORTING.value
    )
    assert response_json["status"] == Status.FAILED.value
    assert response_json["message"] == 'Exception("D\'Oh")'
    assert response_json["exception_type"] == "Exception"


@pytest.mark.timeout(5)
@pytest.mark.parametrize(
    "endpoint, test_file",
    [
        [
            "/hyperion_grid_detect_then_xray_centre/start",
            "tests/test_data/parameter_json_files/good_test_grid_with_edge_detect_parameters.json",
        ],
        [
            "/rotation_scan/start",
            "tests/test_data/parameter_json_files/good_test_one_multi_rotation_scan_parameters.json",
        ],
        [
            "/pin_tip_centre_then_xray_centre/start",
            "tests/test_data/parameter_json_files/good_test_pin_centre_then_xray_centre_parameters.json",
        ],
        [
            "/rotation_scan/start",
            "tests/test_data/parameter_json_files/good_test_multi_rotation_scan_parameters.json",
        ],
        [
            "/load_centre_collect_full/start",
            "tests/test_data/parameter_json_files/good_test_load_centre_collect_params.json",
        ],
    ],
)
def test_start_with_json_file_gives_success(
    test_env: ClientAndRunEngine, endpoint: str, test_file: str, tmp_path: Path
):
    test_env.mock_run_engine.RE_takes_time = False

    test_params = raw_params_from_file(test_file, tmp_path)
    response = test_env.client.put(endpoint, json=test_params)
    check_status_in_response(response, Status.SUCCESS)


@pytest.mark.timeout(3)
def test_start_with_json_file_with_extras_gives_error(
    test_env: ClientAndRunEngine, tmp_path: Path
):
    test_env.mock_run_engine.RE_takes_time = False

    params = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_parameters.json", tmp_path
    )
    params["extra_param"] = "test"
    response = test_env.client.put(START_ENDPOINT, json=params)
    check_status_in_response(response, Status.FAILED)


@pytest.mark.parametrize(
    ["arg_list", "parsed_arg_values"],
    [
        (
            [
                "--dev",
            ],
            (True,),
        ),
        ([], (False,)),
    ],
)
def test_cli_args_parse(arg_list, parsed_arg_values):
    argv[1:] = arg_list
    test_args = parse_cli_args()
    assert test_args.dev_mode == parsed_arg_values[0]


@pytest.mark.skip(
    "Wait for connection doesn't play nice with ophyd-async. See https://github.com/DiamondLightSource/hyperion/issues/1159"
)
def test_when_blueskyrunner_initiated_then_plans_are_setup_and_devices_connected():
    zebra = MagicMock(spec=Zebra)
    attenuator = MagicMock(spec=BinaryFilterAttenuator)

    context = BlueskyContext()
    context.register_device(zebra, "zebra")
    context.register_device(attenuator, "attenuator")

    @dataclass
    class FakeComposite:
        attenuator: BinaryFilterAttenuator
        zebra: Zebra

    # A fake setup for a plan that uses two devices: attenuator and zebra.
    def fake_create_devices(context) -> FakeComposite:
        print("CREATING DEVICES")
        return device_composite_from_context(context, FakeComposite)

    with patch.dict(
        "mx_bluesky.hyperion.__main__.PLAN_REGISTRY",
        {
            "flyscan_xray_centre": {
                "setup": fake_create_devices,
                "param_type": MagicMock(),
            },
        },
        clear=True,
    ):
        print(PLAN_REGISTRY)

        GDARunner(
            context=context,
        )

    zebra.wait_for_connection.assert_called()
    attenuator.wait_for_connection.assert_called()


@patch(
    "mx_bluesky.hyperion.experiment_plans.rotation_scan_plan.create_devices",
    autospec=True,
)
def test_when_blueskyrunner_initiated_then_setup_called_upon_start(
    mock_setup, hyperion_fgs_params: HyperionSpecifiedThreeDGridScan
):
    mock_setup = MagicMock()
    with patch.dict(
        "mx_bluesky.hyperion.__main__.PLAN_REGISTRY",
        {
            "multi_rotation_scan": {
                "setup": mock_setup,
                "param_type": MagicMock(),
            },
        },
        clear=True,
    ):
        runner = GDARunner(MagicMock())
        mock_setup.assert_not_called()
        runner.start(lambda: None, hyperion_fgs_params, "multi_rotation_scan")
        mock_setup.assert_called_once()
        runner.shutdown()


def test_log_on_invalid_json_params(test_env: ClientAndRunEngine):
    test_env.mock_run_engine.RE_takes_time = False
    response = test_env.client.put(TEST_BAD_PARAM_ENDPOINT, data='{"bad":1}').json
    assert isinstance(response, dict)
    assert response.get("status") == Status.FAILED.value
    assert (message := response.get("message")) is not None
    assert message.startswith(
        "ValueError('Supplied parameters don\\'t match the plan for this endpoint"
    )
    assert response.get("exception_type") == "ValueError"


@pytest.mark.timeout(2)
def test_warn_exception_during_plan_causes_warning_in_log(
    caplog: pytest.LogCaptureFixture, test_env: ClientAndRunEngine, test_params
):
    test_env.client.put(START_ENDPOINT, data=test_params)
    test_env.mock_run_engine.error = WarningException(_MULTILINE_MESSAGE)
    response_json = wait_for_run_engine_status(test_env.client)
    assert response_json["status"] == Status.FAILED.value
    assert response_json["message"] == repr(test_env.mock_run_engine.error)
    assert response_json["exception_type"] == "WarningException"
    log_record = [r for r in caplog.records if r.funcName == "wait_on_queue"][0]
    assert log_record.levelname == "WARNING" and _MULTILINE_MESSAGE in getattr(
        log_record, "exc_text", ""
    )


def _raise_exception(*args, **kwargs):
    raise WarningException(_MULTILINE_MESSAGE)


@patch.dict(
    "mx_bluesky.hyperion.__main__.PLAN_REGISTRY",
    {"pin_tip_centre_then_xray_centre": {"param_type": _raise_exception}},
)
def test_exception_during_parameter_decode_generates_nicely_formatted_log_message(
    caplog: pytest.LogCaptureFixture, test_env: ClientAndRunEngine, test_params
):
    response = test_env.client.put(START_ENDPOINT, data=test_params)
    assert response.json["status"] == Status.FAILED.value  # type: ignore
    logrecord = [
        r for r in caplog.records if r.funcName == "put" and r.filename == "__main__.py"
    ][0]
    assert logrecord.levelname == "ERROR"
    assert _MULTILINE_MESSAGE in logrecord.message


@pytest.mark.parametrize("dev_mode", [True, False])
@patch(
    "dodal.devices.i03.undulator_dcm.get_beamline_parameters",
    return_value={"DCM_Perp_Offset_FIXED": 111},
)
def test_when_context_created_then_contains_expected_number_of_plans(
    get_beamline_parameters, dev_mode
):
    from dodal.beamlines import i03

    mock_beamline_module_filepaths("i03", i03)

    with patch.dict(
        os.environ,
        {"BEAMLINE": "i03"},
    ):
        with patch(
            "mx_bluesky.hyperion.utils.context.BlueskyContext.with_dodal_module"
        ) as mock_with_dodal_module:
            context = setup_context(dev_mode=dev_mode)
            mock_with_dodal_module.assert_called_once_with(ANY, mock=dev_mode)
        plan_names = context.plans.keys()

        # assert "rotation_scan" in plan_names
        # May want to add back in if we change name of multi_rotation_scan to rotation_scan
        assert "hyperion_grid_detect_then_xray_centre" in plan_names
        assert "rotation_scan" in plan_names
        assert "pin_tip_centre_then_xray_centre" in plan_names


@pytest.mark.parametrize("dev_mode", [False, True])
def test_context_created_with_dev_mode(dev_mode: bool, mock_setup_context: MagicMock):
    with (
        patch("sys.argv", new=["hyperion", "--dev"] if dev_mode else ["hyperion"]),
        patch("mx_bluesky.hyperion.__main__.create_app"),
        patch("mx_bluesky.hyperion.__main__.GDARunner.wait_on_queue"),
    ):
        main()

    mock_setup_context.assert_called_once_with(dev_mode=dev_mode)


@patch("mx_bluesky.hyperion.__main__.do_default_logging_setup")
@patch("mx_bluesky.hyperion.__main__.alerting.set_alerting_service")
def test_initialise_configures_logging(
    mock_alerting_setup: MagicMock, mock_logging_setup: MagicMock
):
    args = HyperionArgs(mode=HyperionMode.GDA, dev_mode=True)

    initialise_globals(args)

    mock_logging_setup.assert_called_once_with(
        CONST.LOG_FILE_NAME, CONST.GRAYLOG_PORT, dev_mode=True
    )


@patch("mx_bluesky.hyperion.__main__.do_default_logging_setup")
@patch("mx_bluesky.hyperion.__main__.alerting.set_alerting_service")
def test_initialise_configures_alerting(
    mock_alerting_setup: MagicMock, mock_logging_setup: MagicMock
):
    args = HyperionArgs(mode=HyperionMode.GDA, dev_mode=True)

    initialise_globals(args)

    mock_alerting_setup.assert_called_once()
    assert isinstance(mock_alerting_setup.mock_calls[0].args[0], LoggingAlertService)


@patch("sys.argv", new=["hyperion", "--mode", "udc"])
@patch("mx_bluesky.hyperion.__main__.create_app")
@patch("mx_bluesky.hyperion.__main__.do_default_logging_setup")
def test_hyperion_in_udc_mode_starts_logging(
    mock_do_default_logging_setup: MagicMock,
    mock_create_app: MagicMock,
    mock_setup_context: MagicMock,
):
    main()

    mock_do_default_logging_setup.assert_called_once_with(
        CONST.LOG_FILE_NAME, CONST.GRAYLOG_PORT, dev_mode=False
    )
    mock_create_app.assert_not_called()


@patch("sys.argv", new=["hyperion", "--mode", "udc"])
@patch("mx_bluesky.hyperion.__main__.do_default_logging_setup", MagicMock())
@patch("mx_bluesky.hyperion.__main__.run_forever", MagicMock())
def test_hyperion_in_udc_mode_starts_udc_api(
    mock_create_udc_server: MagicMock,
    mock_setup_context: MagicMock,
):
    main()
    mock_create_udc_server.assert_called_once()
    assert isinstance(mock_create_udc_server.mock_calls[0].args[0], PlanRunner)


@patch("sys.argv", new=["hyperion", "--mode", "udc"])
@patch("mx_bluesky.hyperion.__main__.setup_context")
@patch("mx_bluesky.hyperion.__main__.run_forever")
@patch("mx_bluesky.hyperion.baton_handler.find_device_in_context", autospec=True)
def test_hyperion_in_udc_mode_starts_udc_loop(
    mock_find_device_in_context: MagicMock,
    mock_run_forever: MagicMock,
    mock_setup_context: MagicMock,
    mock_flask_thread: MagicMock,
):
    main()

    mock_run_forever.assert_called_once()
    assert isinstance(mock_run_forever.mock_calls[0].args[0], PlanRunner)


@patch("sys.argv", new=["hyperion", "--mode", "gda"])
@patch("mx_bluesky.hyperion.__main__.setup_context", MagicMock())
@patch("mx_bluesky.hyperion.__main__.GDARunner.wait_on_queue")
def test_hyperion_in_gda_mode_doesnt_start_udc_loop(
    mock_gda_runner: MagicMock,
    mock_flask_thread: MagicMock,
):
    main()

    mock_gda_runner.assert_called_once()


@patch("mx_bluesky.hyperion.__main__.Api")
@patch("mx_bluesky.hyperion.__main__.setup_context", MagicMock())
@patch("mx_bluesky.hyperion.baton_handler.find_device_in_context", MagicMock())
@patch("mx_bluesky.hyperion.runner.GDARunner.wait_on_queue", MagicMock())
@patch("mx_bluesky.hyperion.__main__.run_forever", MagicMock())
@pytest.mark.parametrize("mode", ["gda", "udc"])
def test_hyperion_exposes_run_endpoint_only_if_gda_mode_selected(
    mock_api: MagicMock,
    mock_flask_thread: MagicMock,
    mode: str,
):
    with patch("sys.argv", new=["hyperion", "--mode", mode]):
        main()

    assert len(
        [
            c
            for c in mock_api.return_value.add_resource.mock_calls
            if c.args[0] == RunExperiment
        ]
    ) == (1 if mode == "gda" else 0)


@patch("mx_bluesky.hyperion.__main__.flush_debug_handler")
def test_flush_logs(mock_flush_debug_handler: MagicMock, test_env: ClientAndRunEngine):
    response = test_env.client.put(FLUSH_LOGS_ENDPOINT)
    check_status_in_response(response, Status.SUCCESS)
    mock_flush_debug_handler.assert_called_once()


@patch("sys.argv", new=["hyperion", "--mode", "udc", "--dev"])
@patch(
    "mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon",
    return_value=[AGAMEMNON_WAIT_INSTRUCTION],
)
@patch("mx_bluesky.hyperion.baton_handler.clear_all_device_caches", MagicMock())
@patch("mx_bluesky.hyperion.baton_handler.setup_devices", MagicMock())
def test_sending_main_process_sigterm_in_udc_mode_performs_clean_prompt_shutdown(
    mock_create_parameters_from_agamemnon,
    use_beamline_t01,
    mock_create_udc_server,
):
    def wait_for_udc_to_start_then_send_sigterm():
        while len(mock_create_udc_server.mock_calls) == 0:
            sleep(0.2)

        plan_runner = mock_create_udc_server.mock_calls[0].args[0]
        context = plan_runner.context
        baton = find_device_in_context(context, "baton", Baton)
        set_mock_value(baton.requested_user, HYPERION_USER)
        while len(mock_create_parameters_from_agamemnon.mock_calls) == 0:
            sleep(0.2)
        os.kill(os.getpid(), signal.SIGTERM)

    t = threading.Thread(None, wait_for_udc_to_start_then_send_sigterm, daemon=True)
    t.start()
    main()



import bluesky.plan_stubs as bps
import pydantic
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue
from dodal.devices.collimation_table import CollimationTable
from dodal.devices.cryostream import CryoStream
from dodal.devices.cryostream import InOut as CryoInOut
from dodal.devices.fluorescence_detector_motion import (
    FluorescenceDetector,
)
from dodal.devices.fluorescence_detector_motion import InOut as FlouInOut
from dodal.devices.mx_phase1.beamstop import Beamstop, BeamstopPositions
from dodal.devices.scintillator import InOut as ScinInOut
from dodal.devices.scintillator import Scintillator


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class UDCDefaultDevices:
    cryostream: CryoStream
    fluorescence_det_motion: FluorescenceDetector
    beamstop: Beamstop
    scintillator: Scintillator
    aperture_scatterguard: ApertureScatterguard
    collimation_table: CollimationTable


def move_to_udc_default_state(devices: UDCDefaultDevices):
    """Moves beamline to known positions prior to UDC start"""

    cryostream_temp = yield from bps.rd(devices.cryostream.temperature_k)
    cryostream_pressure = yield from bps.rd(devices.cryostream.back_pressure_bar)
    if cryostream_temp > devices.cryostream.MAX_TEMP_K:
        raise ValueError("Cryostream temperature is too high, not starting UDC")
    if cryostream_pressure > devices.cryostream.MAX_PRESSURE_BAR:
        raise ValueError("Cryostream back pressure is too high, not starting UDC")

    yield from bps.abs_set(devices.scintillator.selected_pos, ScinInOut.OUT, wait=True)

    yield from bps.abs_set(
        devices.fluorescence_det_motion.pos, FlouInOut.OUT, group="udc_default"
    )

    yield from bps.abs_set(devices.collimation_table.inboard_y, 0, group="udc_default")
    yield from bps.abs_set(devices.collimation_table.outboard_y, 0, group="udc_default")
    yield from bps.abs_set(devices.collimation_table.upstream_y, 0, group="udc_default")
    yield from bps.abs_set(devices.collimation_table.upstream_x, 0, group="udc_default")
    yield from bps.abs_set(
        devices.collimation_table.downstream_x, 0, group="udc_default"
    )

    yield from bps.abs_set(
        devices.beamstop.selected_pos,
        BeamstopPositions.DATA_COLLECTION,
        group="udc_default",
    )

    yield from bps.abs_set(
        devices.aperture_scatterguard.selected_aperture,
        ApertureValue.SMALL,
        group="udc_default",
    )

    yield from bps.abs_set(devices.cryostream.course, CryoInOut.IN, group="udc_default")
    yield from bps.abs_set(devices.cryostream.fine, CryoInOut.IN, group="udc_default")

    yield from bps.wait("udc_default")



from __future__ import annotations

from unittest.mock import MagicMock

import pytest
from bluesky.run_engine import RunEngine
from dodal.devices.aperturescatterguard import ApertureScatterguard
from dodal.devices.motors import XYZStage
from dodal.devices.robot import BartRobot
from dodal.devices.smargon import Smargon
from ophyd_async.testing import set_mock_value
from requests import get
from tests.conftest import SimConstants

from mx_bluesky.common.device_setup_plans.robot_load_unload import robot_unload
from mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback import (
    RobotLoadISPyBCallback,
)


@pytest.mark.system_test
def test_execute_unload_sample_full(
    RE: RunEngine,
    robot: BartRobot,
    smargon: Smargon,
    aperture_scatterguard: ApertureScatterguard,
    lower_gonio: XYZStage,
):
    callback = RobotLoadISPyBCallback()
    old_start_robot_action = callback.expeye.start_robot_action
    action_id = None

    def get_robot_action_id(*args, **kwargs):
        nonlocal action_id
        action_id = old_start_robot_action(*args, **kwargs)
        return action_id

    callback.expeye.start_robot_action = MagicMock(side_effect=get_robot_action_id)

    set_mock_value(robot.sample_id, SimConstants.ST_SAMPLE_ID)
    set_mock_value(robot.current_puck, 10)

    RE.subscribe(callback)
    RE(
        robot_unload(
            robot, smargon, aperture_scatterguard, lower_gonio, SimConstants.ST_VISIT
        )
    )
    get_robot_data_url = f"{callback.expeye._base_url}/robot-actions/{action_id}"
    response = get(get_robot_data_url, auth=callback.expeye._auth)

    assert response.ok
    response = response.json()
    assert response["robotActionId"] == action_id
    assert response["status"] == "SUCCESS"
    assert response["sampleId"] == SimConstants.ST_SAMPLE_ID
    assert response["dewarLocation"] == 10
    assert response["actionType"] == "UNLOAD"
    assert response["message"] == "OK"



import json
from dataclasses import fields
from enum import Enum
from pathlib import Path
from typing import Any, Generic, TypeVar

from daq_config_server.client import ConfigServer
from pydantic import TypeAdapter

from mx_bluesky.common.parameters.constants import (
    GDA_DOMAIN_PROPERTIES_PATH,
    FeatureSetting,
    FeatureSettingources,
    OavConstants,
)
from mx_bluesky.common.utils.log import LOGGER

FEATURE_FLAG_CACHE_LENGTH_S = 60 * 5
# Used by the config server when refreshing its cache
_JSON_CONFIG_PATHS = [OavConstants.OAV_CONFIG_JSON]

T = TypeVar("T", bound=FeatureSetting)


class MXConfigClient(ConfigServer, Generic[T]):
    def __init__(
        self,
        feature_sources: type[FeatureSettingources],
        feature_dc: type[T],
        url: str = "https://daq-config.diamond.ac.uk",
    ):
        """MX implementation of the config server client. Makes requests to the config server to retrieve config while falling back to
        the filesystem in the case that the request failed.

        See mx_bluesky/hyperion/external_interaction/config_server.py for example implementation.

        Args:
        feature_sources: A StrEnum containing available features, where the string is the name of that feature toggle in a beamline's GDA
        domain.properties.

        feature_dc: A dataclass containing available features along with their default flags. This dataclass must contain the same keys
        as the feature_sources parameter. These defaults are used when a server request fails.
        """

        self.feature_sources = feature_sources
        self.feature_dc: type[T] = feature_dc
        self._cached_features: T | None = None
        self._cached_json_config: dict[str, dict[str, Any]] = {}
        self._time_of_last_feature_get: float = 0
        self._verify_feature_parameters()
        super().__init__(url)

    def get_feature_flags(self) -> T:
        """Get feature flags by making a request to the config server. If the request fails, use the hardcoded defaults. Store results in a cache
        which should be updated at the start of a plan using self.refresh_cache()
        """
        return self._get_feature_flags()

    def get_json_config(self, path_to_json: Path | str) -> dict[str, Any]:
        """Get the OAV config in the form of a python dictionary. Store results in a cache
        which should be updated at the start of a plan using self.refresh_cache().

        Note that this method replicates the functionality of the config server, but provides
        a fallback incase of server. It will be removed once the config server has been run reliably
        for several months.

        Args:
        path_to_json: Absolute file path to json file
        """
        return self._get_json_config(path_to_json)

    def refresh_cache(self):
        """Refresh the client's cache. Use at the beginning of a plan"""
        self._get_feature_flags(reset_cached_result=True)
        for path in _JSON_CONFIG_PATHS:
            self._get_json_config(path, reset_cached_result=True)

    def _verify_feature_parameters(self):
        sources_keys = {feature.name for feature in self.feature_sources}
        feature_dc_keys = {key.name for key in fields(self.feature_dc)}
        assert sources_keys == feature_dc_keys, (
            f"MXConfig server feature_sources names do not match feature_dc keys: {sources_keys} != {feature_dc_keys}"
        )

    def _get_feature_flags(self, reset_cached_result=False) -> T:
        """
        Args:
        reset_cached_result (bool): Force refresh the cache for this request
        """
        try:
            if reset_cached_result:
                self._cached_features = None
            if not self._cached_features:
                self._cached_features = None
                # Construct self.feature_dc by reading the domain.properties file
                all_features = list(self.feature_sources)
                feature_dict = {}
                domain_properties = self.get_file_contents(
                    GDA_DOMAIN_PROPERTIES_PATH, reset_cached_result=reset_cached_result
                ).splitlines()
                for line in domain_properties:
                    line = line.strip()
                    if not line or line.startswith("#"):
                        continue
                    line = line.split("#", 1)[0].strip()  # Remove inline comments
                    if "=" in line:
                        key, value = map(str.strip, line.split("=", 1))
                        for feature in all_features:
                            assert isinstance(feature, Enum)
                            if key == feature.value:
                                feature_dict[feature.name] = value
                self._check_missing_fields(
                    {f.name for f in fields(self.feature_dc)}, set(feature_dict.keys())
                )
                self._cached_features = self.feature_dc(**feature_dict)
            return self._cached_features
        except Exception as e:
            LOGGER.warning(
                f"Failed to get feature flags from config server: {e} \nUsing defaults..."
            )
            return self.feature_dc()

    def _check_missing_fields(self, expected: set, actual: set):
        missing = expected - actual
        if missing:
            LOGGER.warning(
                f"Missing features from domain.properties: {missing}.\n Using defaults for missing features"
            )

    def _get_json_config(
        self, path_to_json: Path | str, reset_cached_result=False
    ) -> dict[str, Any]:
        """
        Args:
        reset_cached_result (bool): Force refresh the cache for this request
        """
        if reset_cached_result:
            self._cached_json_config = {}
        str_to_json = str(path_to_json)
        if str_to_json not in self._cached_json_config:
            try:
                self._cached_json_config[str_to_json] = self.get_file_contents(
                    path_to_json, dict, reset_cached_result=reset_cached_result
                )

            except Exception as e:
                LOGGER.warning(
                    f"Failed to get json config from config server: {e} \nReading the file directly..."
                )
                with open(path_to_json) as f:
                    self._cached_json_config[str_to_json] = TypeAdapter(
                        dict[str, Any]
                    ).validate_python(json.loads(f.read()))
        return self._cached_json_config[str_to_json]



from typing import Protocol, runtime_checkable

import bluesky.plan_stubs as bps
from bluesky.utils import MsgGenerator
from dodal.devices.zebra.zebra import (
    Zebra,
)
from dodal.devices.zebra.zebra_controlled_shutter import (
    ZebraShutter,
    ZebraShutterControl,
)

from mx_bluesky.common.parameters.constants import ZEBRA_STATUS_TIMEOUT
from mx_bluesky.common.utils.log import LOGGER


@runtime_checkable
class GridscanSetupDevices(Protocol):
    zebra: Zebra
    sample_shutter: ZebraShutter


def setup_zebra_for_gridscan(
    composite: GridscanSetupDevices,  # XRC gridscan's generic trigger setup expects a composite rather than individual devices
    group="setup_zebra_for_gridscan",
    wait=True,
) -> MsgGenerator:
    zebra = composite.zebra
    # Set shutter to automatic and to trigger via motion controller GPIO signal (IN4_TTL)
    yield from configure_zebra_and_shutter_for_auto_shutter(
        zebra, composite.sample_shutter, zebra.mapping.sources.IN4_TTL, group=group
    )

    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_DETECTOR],
        zebra.mapping.sources.IN3_TTL,
        group=group,
    )
    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_XSPRESS3],
        zebra.mapping.sources.DISCONNECT,
        group=group,
    )
    yield from bps.abs_set(
        zebra.output.pulse_1.input, zebra.mapping.sources.DISCONNECT, group=group
    )

    if wait:
        yield from bps.wait(group, timeout=ZEBRA_STATUS_TIMEOUT)


def set_shutter_auto_input(zebra: Zebra, input: int, group="set_shutter_trigger"):
    """Set the signal that controls the shutter. We use the second input to the
    Zebra's AND2 gate for this input. ZebraShutter control mode must be in auto for this input to take control

    For more details see the ZebraShutter device."""
    auto_gate = zebra.mapping.AND_GATE_FOR_AUTO_SHUTTER
    auto_shutter_control = zebra.logic_gates.and_gates[auto_gate]
    yield from bps.abs_set(auto_shutter_control.sources[2], input, group)


def configure_zebra_and_shutter_for_auto_shutter(
    zebra: Zebra, zebra_shutter: ZebraShutter, input: int, group="use_automatic_shutter"
):
    """Set the shutter to auto mode, and configure the zebra to trigger the shutter on
    an input source. For the input, use one of the source constants in zebra.py

    When the shutter is in auto/manual, logic in EPICS sets the Zebra's
    SOFT_IN1 to low/high respectively. The Zebra's AND2 gate should be used to control the shutter while in auto mode.
    To do this, we need (AND2 = SOFT_IN1 AND input), where input is the zebra signal we want to control the shutter when in auto mode.
    """
    # See https://github.com/DiamondLightSource/dodal/issues/813 for better typing here.

    # Set shutter to auto mode
    yield from bps.abs_set(
        zebra_shutter.control_mode, ZebraShutterControl.AUTO, group=group
    )

    auto_gate = zebra.mapping.AND_GATE_FOR_AUTO_SHUTTER

    # Set first input of AND2 gate to SOFT_IN1, which is high when shutter is in auto mode
    # Note the Zebra should ALWAYS be setup this way. See https://github.com/DiamondLightSource/mx-bluesky/issues/551
    yield from bps.abs_set(
        zebra.logic_gates.and_gates[auto_gate].sources[1],
        zebra.mapping.sources.SOFT_IN1,
        group=group,
    )

    # Set the second input of AND2 gate to the requested zebra input source
    yield from set_shutter_auto_input(zebra, input, group=group)


def tidy_up_zebra_after_gridscan(
    zebra: Zebra,
    zebra_shutter: ZebraShutter,
    group="tidy_up_zebra_after_gridscan",
    wait=True,
) -> MsgGenerator:
    LOGGER.info("Tidying up Zebra")

    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_DETECTOR],
        zebra.mapping.sources.PC_PULSE,
        group=group,
    )
    yield from bps.abs_set(
        zebra_shutter.control_mode, ZebraShutterControl.MANUAL, group=group
    )
    yield from set_shutter_auto_input(zebra, zebra.mapping.sources.PC_GATE, group=group)

    if wait:
        yield from bps.wait(group, timeout=ZEBRA_STATUS_TIMEOUT)



import pydantic
from dodal.devices.aperturescatterguard import (
    ApertureScatterguard,
)
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.backlight import Backlight
from dodal.devices.common_dcm import BaseDCM
from dodal.devices.detector.detector_motion import DetectorMotion
from dodal.devices.eiger import EigerDetector
from dodal.devices.fast_grid_scan import (
    ZebraFastGridScan,
)
from dodal.devices.flux import Flux
from dodal.devices.mx_phase1.beamstop import Beamstop
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.robot import BartRobot
from dodal.devices.s4_slit_gaps import S4SlitGaps
from dodal.devices.smargon import Smargon
from dodal.devices.synchrotron import Synchrotron
from dodal.devices.undulator import Undulator
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.devices.zebra.zebra import Zebra
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutter
from dodal.devices.zocalo import ZocaloResults


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class FlyScanEssentialDevices:
    eiger: EigerDetector
    synchrotron: Synchrotron
    zocalo: ZocaloResults
    smargon: Smargon


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class OavGridDetectionComposite:
    """All devices which are directly or indirectly required by this plan"""

    backlight: Backlight
    oav: OAV
    smargon: Smargon
    pin_tip_detection: PinTipDetection


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class GridDetectThenXRayCentreComposite(FlyScanEssentialDevices):
    """All devices which are directly or indirectly required by this plan"""

    aperture_scatterguard: ApertureScatterguard
    attenuator: BinaryFilterAttenuator
    backlight: Backlight
    beamstop: Beamstop
    dcm: BaseDCM
    detector_motion: DetectorMotion
    zebra_fast_grid_scan: ZebraFastGridScan
    flux: Flux
    oav: OAV
    pin_tip_detection: PinTipDetection
    s4_slit_gaps: S4SlitGaps
    undulator: Undulator
    xbpm_feedback: XBPMFeedback
    zebra: Zebra
    robot: BartRobot
    sample_shutter: ZebraShutter



import logging
import os
from unittest.mock import MagicMock, patch

import pytest
from bluesky import RunEngine
from bluesky import plan_stubs as bps
from bluesky.preprocessors import run_decorator
from dodal.log import get_graylog_configuration, set_up_graylog_handler

from mx_bluesky.common.external_interaction.alerting import (
    Metadata,
    get_alerting_service,
    set_alerting_service,
)
from mx_bluesky.common.external_interaction.alerting.log_based_service import (
    LoggingAlertService,
)
from mx_bluesky.common.external_interaction.callbacks.sample_handling.sample_handling_callback import (
    SampleHandlingCallback,
)
from mx_bluesky.hyperion.baton_handler import (
    _raise_baton_released_alert,
    _raise_udc_completed_alert,
    _raise_udc_start_alert,
)
from mx_bluesky.hyperion.parameters.constants import CONST, HyperionConstants

from .....conftest import SimConstants


@pytest.fixture(autouse=True)
def setup_graylog():
    logger = logging.getLogger("Dodal")
    host, _ = get_graylog_configuration(False)
    set_up_graylog_handler(logger, host, HyperionConstants.GRAYLOG_PORT)


@pytest.fixture
def patch_raise_alert_to_disable_ehc_notifications():
    """Patch raise_alert so that TEST is prepended to the alert summary to avoid downstream
    email filters from forwarding to the EHC"""

    def patched_raise_alert(summary: str, content: str, metadata: dict[Metadata, str]):
        return get_alerting_service().raise_alert(
            "TEST Unicorn Defence Commission Excursion", content, metadata
        )

    with patch(
        "mx_bluesky.common.external_interaction.callbacks.sample_handling.sample_handling_callback"
        ".get_alerting_service",
        return_value=MagicMock(**{"raise_alert.side_effect": patched_raise_alert}),  # type: ignore
    ):
        yield


@pytest.mark.requires(external="graylog")
def test_alert_to_graylog():
    alert_service = LoggingAlertService(CONST.GRAYLOG_STREAM_ID)
    alert_service.raise_alert(
        "Test alert", "This is a test.", {Metadata.SAMPLE_ID: "12345"}
    )


@patch(
    "mx_bluesky.common.external_interaction.callbacks.sample_handling.sample_handling_callback.ExpeyeInteraction",
    MagicMock(),
)
@pytest.mark.requires(external="graylog")
@patch.dict(os.environ, {"BEAMLINE": "i03"})
def test_alert_from_plan_exception(
    RE: RunEngine, patch_raise_alert_to_disable_ehc_notifications
):
    RE.subscribe(SampleHandlingCallback())
    set_alerting_service(LoggingAlertService(CONST.GRAYLOG_STREAM_ID))

    @run_decorator(
        md={
            "activate_callbacks": ["SampleHandlingCallback"],
            "metadata": {
                "sample_id": SimConstants.ST_SAMPLE_ID,
                "visit": SimConstants.ST_VISIT,
                "container": SimConstants.ST_CONTAINER_ID,
            },
        }
    )
    def plan_with_exception():
        yield from bps.null()
        raise RuntimeError("Test exception.")

    with pytest.raises(RuntimeError):
        RE(plan_with_exception())


@pytest.mark.requires(external="graylog")
@patch.dict(os.environ, {"BEAMLINE": "i03"})
def test_alert_udc_start():
    service = LoggingAlertService(CONST.GRAYLOG_STREAM_ID)
    _raise_udc_start_alert(service)


@pytest.mark.requires(external="graylog")
@patch.dict(os.environ, {"BEAMLINE": "i03"})
def test_alert_udc_baton_released():
    service = LoggingAlertService(CONST.GRAYLOG_STREAM_ID)
    _raise_baton_released_alert(service, "GDA")


@pytest.mark.requires(external="graylog")
@patch.dict(os.environ, {"BEAMLINE": "i03"})
def test_alert_udc_completed():
    service = LoggingAlertService(CONST.GRAYLOG_STREAM_ID)
    _raise_udc_completed_alert(service)



from mx_bluesky.common.parameters.components import MxBlueskyParameters


class Wait(MxBlueskyParameters):
    """Represents an instruction from Agamemnon for Hyperion to wait for a specified time
    Attributes:
        duration_s: duration to wait in seconds
    """

    duration_s: float



import asyncio
import pprint
import sys
import time
from collections.abc import Callable
from functools import partial
from typing import cast
from unittest.mock import MagicMock, patch

import pytest
from _pytest.fixtures import FixtureRequest
from bluesky.run_engine import RunEngine
from dodal.beamlines import i03
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue
from dodal.devices.backlight import Backlight
from dodal.devices.detector.detector_motion import DetectorMotion
from dodal.devices.eiger import EigerDetector
from dodal.devices.fast_grid_scan import PandAFastGridScan, ZebraFastGridScan
from dodal.devices.flux import Flux
from dodal.devices.i03 import Beamstop
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.robot import BartRobot
from dodal.devices.s4_slit_gaps import S4SlitGaps
from dodal.devices.smargon import Smargon
from dodal.devices.synchrotron import Synchrotron, SynchrotronMode
from dodal.devices.zocalo import ZocaloResults, ZocaloTrigger
from event_model.documents import Event
from ophyd_async.core import AsyncStatus
from ophyd_async.fastcs.panda import HDFPanda
from ophyd_async.testing import set_mock_value

from mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan import (
    BeamlineSpecificFGSFeatures,
    FlyScanEssentialDevices,
)
from mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback import (
    ZocaloCallback,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    GridscanISPyBCallback,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.nexus_callback import (
    GridscanNexusFileCallback,
)
from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionGroupInfo,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
    StoreInIspyb,
)
from mx_bluesky.common.parameters.constants import (
    DocDescriptorNames,
    EnvironmentConstants,
    PlanNameConstants,
)
from mx_bluesky.common.parameters.device_composites import (
    GridDetectThenXRayCentreComposite,
)
from mx_bluesky.common.parameters.gridscan import GridCommon, SpecifiedThreeDGridScan
from mx_bluesky.hyperion.parameters.device_composites import (
    HyperionGridDetectThenXRayCentreComposite,
)
from tests.conftest import raw_params_from_file


@pytest.fixture
async def RE():
    RE = RunEngine(call_returns_result=True)
    # make sure the event loop is thoroughly up and running before we try to create
    # any ophyd_async devices which might need it
    timeout = time.monotonic() + 1
    while not RE.loop.is_running():
        await asyncio.sleep(0)
        if time.monotonic() > timeout:
            raise TimeoutError("This really shouldn't happen but just in case...")
    yield RE
    # RunEngine creates its own loop if we did not supply it, we must terminate it
    RE.loop.call_soon_threadsafe(RE.loop.stop)
    RE._th.join()


_ALLOWED_PYTEST_TASKS = {"async_finalizer", "async_setup", "async_teardown"}


def _error_and_kill_pending_tasks(
    loop: asyncio.AbstractEventLoop, test_name: str, test_passed: bool
) -> set[asyncio.Task]:
    """Cancels pending tasks in the event loop for a test. Raises an exception if
    the test hasn't already.

    Args:
        loop: The event loop to check for pending tasks.
        test_name: The name of the test.
        test_passed: Indicates whether the test passed.

    Returns:
        set[asyncio.Task]: The set of unfinished tasks that were cancelled.

    Raises:
        RuntimeError: If there are unfinished tasks and the test didn't fail.
    """
    unfinished_tasks = {
        task
        for task in asyncio.all_tasks(loop)
        if (coro := task.get_coro()) is not None
        and hasattr(coro, "__name__")
        and coro.__name__ not in _ALLOWED_PYTEST_TASKS
        and not task.done()
    }
    for task in unfinished_tasks:
        task.cancel()

    # We only raise an exception here if the test didn't fail anyway.
    # If it did then it makes sense that there's some tasks we need to cancel,
    # but an exception will already have been raised.
    if unfinished_tasks and test_passed:
        raise RuntimeError(
            f"Not all tasks closed during test {test_name}:\n"
            f"{pprint.pformat(unfinished_tasks, width=88)}"
        )

    return unfinished_tasks


@pytest.fixture(autouse=True, scope="function")
async def fail_test_on_unclosed_tasks(request: FixtureRequest):
    """
    Used on every test to ensure failure if there are pending tasks
    by the end of the test.
    """

    fail_count = request.session.testsfailed
    loop = asyncio.get_running_loop()

    loop.set_debug(True)

    request.addfinalizer(
        lambda: _error_and_kill_pending_tasks(
            loop, request.node.name, request.session.testsfailed == fail_count
        )
    )


BASIC_PRE_SETUP_DOC = {
    "undulator-current_gap": 0,
    "synchrotron-synchrotron_mode": SynchrotronMode.USER,
    "s4_slit_gaps-xgap": 0,
    "s4_slit_gaps-ygap": 0,
    "smargon-x": 10.0,
    "smargon-y": 20.0,
    "smargon-z": 30.0,
}

BASIC_POST_SETUP_DOC = {
    "aperture_scatterguard-selected_aperture": ApertureValue.OUT_OF_BEAM,
    "aperture_scatterguard-radius": None,
    "aperture_scatterguard-aperture-x": 15,
    "aperture_scatterguard-aperture-y": 16,
    "aperture_scatterguard-aperture-z": 2,
    "aperture_scatterguard-scatterguard-x": 18,
    "aperture_scatterguard-scatterguard-y": 19,
    "attenuator-actual_transmission": 0,
    "flux-flux_reading": 10,
    "dcm-energy_in_kev": 11.105,
}


def assert_event(mock_call, expected):
    actual = mock_call.args[0]
    if "data" in actual:
        actual = actual["data"]
    for k, v in expected.items():
        assert actual[k] == v, f"Mismatch in key {k}, {actual} <=> {expected}"


def create_gridscan_callbacks() -> tuple[
    GridscanNexusFileCallback, GridscanISPyBCallback
]:
    return (
        GridscanNexusFileCallback(param_type=SpecifiedThreeDGridScan),
        GridscanISPyBCallback(
            param_type=SpecifiedThreeDGridScan,
            emit=ZocaloCallback(
                PlanNameConstants.DO_FGS, EnvironmentConstants.ZOCALO_ENV
            ),
        ),
    )


@pytest.fixture
def use_beamline_t01():
    """Beamline t01 is a beamline for unit tests that just contains a baton, so that
    loading the beamline context does not require importing lots of modules and instantiating
    many devices"""
    with patch.dict("os.environ", {"BEAMLINE": "t01"}):
        import tests.unit_tests.t01

        with (
            patch.dict(sys.modules, {"dodal.beamlines.t01": tests.unit_tests.t01}),
            patch("mx_bluesky.hyperion.baton_handler.move_to_udc_default_state"),
            patch("mx_bluesky.hyperion.baton_handler.device_composite_from_context"),
        ):
            yield


@pytest.fixture
def mock_subscriptions(test_fgs_params):
    with (
        patch(
            "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger",
            modified_interactor_mock,
        ),
        patch(
            "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.StoreInIspyb.append_to_comment"
        ),
        patch(
            "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.StoreInIspyb.begin_deposition",
            new=MagicMock(
                return_value=IspybIds(
                    data_collection_ids=(0, 0), data_collection_group_id=0
                )
            ),
        ),
        patch(
            "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.StoreInIspyb.update_deposition",
            new=MagicMock(
                return_value=IspybIds(
                    data_collection_ids=(0, 0),
                    data_collection_group_id=0,
                    grid_ids=(0, 0),
                )
            ),
        ),
    ):
        nexus_callback, ispyb_callback = create_gridscan_callbacks()
        ispyb_callback.ispyb = MagicMock(spec=StoreInIspyb)

    return (nexus_callback, ispyb_callback)


@pytest.fixture
def RE_with_subs(
    RE: RunEngine,
    mock_subscriptions: tuple[GridscanNexusFileCallback | GridscanISPyBCallback],
):
    for cb in list(mock_subscriptions):
        RE.subscribe(cb)
    yield RE, mock_subscriptions


@pytest.fixture
def test_fgs_params(tmp_path):
    return SpecifiedThreeDGridScan(
        **raw_params_from_file(
            "tests/test_data/parameter_json_files/good_test_parameters.json", tmp_path
        )
    )


def mock_zocalo_trigger(zocalo: ZocaloResults, result):
    @AsyncStatus.wrap
    async def mock_complete(results):
        await zocalo._put_results(results, {"dcid": 0, "dcgid": 0})

    zocalo.trigger = MagicMock(side_effect=partial(mock_complete, result))


def modified_interactor_mock(assign_run_end: Callable | None = None):
    mock = MagicMock(spec=ZocaloTrigger)
    if assign_run_end:
        mock.run_end = assign_run_end
    return mock


def modified_store_grid_scan_mock(*args, dcids=(0, 0), dcgid=0, **kwargs):
    mock = MagicMock(spec=StoreInIspyb)
    mock.begin_deposition.return_value = IspybIds(
        data_collection_ids=dcids, data_collection_group_id=dcgid
    )
    mock.update_deposition.return_value = IspybIds(
        data_collection_ids=dcids, data_collection_group_id=dcgid, grid_ids=(0, 0)
    )
    return mock


def make_event_doc(data, descriptor="abc123") -> Event:
    return {
        "time": 0,
        "timestamps": {"a": 0},
        "seq_num": 0,
        "uid": "not so random uid",
        "descriptor": descriptor,
        "data": data,
    }


def run_generic_ispyb_handler_setup(
    ispyb_handler: GridscanISPyBCallback,
    params: SpecifiedThreeDGridScan,
):
    """This is useful when testing 'run_gridscan_and_move(...)' because this stuff
    happens at the start of the outer plan."""

    ispyb_handler.active = True
    ispyb_handler.activity_gated_start(
        {
            "subplan_name": PlanNameConstants.GRIDSCAN_OUTER,
            "mx_bluesky_parameters": params.model_dump_json(),
        }  # type: ignore
    )
    ispyb_handler.activity_gated_descriptor(
        {"uid": "123abc", "name": DocDescriptorNames.HARDWARE_READ_PRE}  # type: ignore
    )
    ispyb_handler.activity_gated_event(
        make_event_doc(
            BASIC_PRE_SETUP_DOC,
            descriptor="123abc",
        )
    )
    ispyb_handler.activity_gated_descriptor(
        {"uid": "abc123", "name": DocDescriptorNames.HARDWARE_READ_DURING}  # type: ignore
    )
    ispyb_handler.activity_gated_event(
        make_event_doc(
            BASIC_POST_SETUP_DOC,
            descriptor="abc123",
        )
    )


@pytest.fixture
async def zebra_fast_grid_scan():
    zebra_fast_grid_scan = i03.zebra_fast_grid_scan(connect_immediately=True, mock=True)
    set_mock_value(zebra_fast_grid_scan.scan_invalid, False)
    set_mock_value(zebra_fast_grid_scan.position_counter, 0)
    return zebra_fast_grid_scan


@pytest.fixture
async def fake_fgs_composite(
    smargon: Smargon,
    test_fgs_params: SpecifiedThreeDGridScan,
    RE: RunEngine,
    done_status,
    attenuator,
    xbpm_feedback,
    synchrotron,
    aperture_scatterguard,
    zocalo,
    panda,
    backlight,
):
    fake_composite = FlyScanEssentialDevices(
        # We don't use the eiger fixture here because .unstage() is used in some tests
        eiger=i03.eiger(connect_immediately=True, mock=True),
        smargon=smargon,
        synchrotron=synchrotron,
        zocalo=zocalo,
    )

    fake_composite.eiger.stage = MagicMock(return_value=done_status)
    # unstage should be mocked on a per-test basis because several rely on unstage
    fake_composite.eiger.set_detector_parameters(test_fgs_params.detector_params)
    fake_composite.eiger.stop_odin_when_all_frames_collected = MagicMock()
    fake_composite.eiger.odin.check_and_wait_for_odin_state = lambda timeout: True

    test_result = {
        "centre_of_mass": [6, 6, 6],
        "max_voxel": [5, 5, 5],
        "max_count": 123456,
        "n_voxels": 321,
        "total_count": 999999,
        "bounding_box": [[3, 3, 3], [9, 9, 9]],
        "sample_id": 12345,
    }

    @AsyncStatus.wrap
    async def mock_complete(result):
        await fake_composite.zocalo._put_results([result], {"dcid": 0, "dcgid": 0})

    fake_composite.zocalo.trigger = MagicMock(
        side_effect=partial(mock_complete, test_result)
    )  # type: ignore
    fake_composite.zocalo.timeout_s = 3
    set_mock_value(fake_composite.smargon.x.max_velocity, 10)

    return fake_composite


@pytest.fixture
def dummy_rotation_data_collection_group_info():
    return DataCollectionGroupInfo(
        visit_string="cm31105-4",
        experiment_type="SAD",
        sample_id=364758,
    )


@pytest.fixture
def beamline_specific(
    zebra_fast_grid_scan: ZebraFastGridScan,
) -> BeamlineSpecificFGSFeatures:
    return BeamlineSpecificFGSFeatures(
        setup_trigger_plan=MagicMock(),
        tidy_plan=MagicMock(),
        set_flyscan_params_plan=MagicMock(),
        fgs_motors=zebra_fast_grid_scan,
        read_pre_flyscan_plan=MagicMock(),
        read_during_collection_plan=MagicMock(),
        get_xrc_results_from_zocalo=False,
    )


@pytest.fixture
def test_full_grid_scan_params(tmp_path):
    params = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_grid_with_edge_detect_parameters.json",
        tmp_path,
    )
    return GridCommon(**params)


@pytest.fixture
async def grid_detect_xrc_devices(
    aperture_scatterguard: ApertureScatterguard,
    backlight: Backlight,
    beamstop_phase1: Beamstop,
    detector_motion: DetectorMotion,
    eiger: EigerDetector,
    smargon: Smargon,
    oav: OAV,
    ophyd_pin_tip_detection: PinTipDetection,
    zocalo: ZocaloResults,
    synchrotron: Synchrotron,
    fast_grid_scan: ZebraFastGridScan,
    s4_slit_gaps: S4SlitGaps,
    flux: Flux,
    zebra,
    zebra_shutter,
    xbpm_feedback,
    attenuator,
    undulator,
    dcm,
):
    yield GridDetectThenXRayCentreComposite(
        aperture_scatterguard=aperture_scatterguard,
        attenuator=attenuator,
        backlight=backlight,
        beamstop=beamstop_phase1,
        detector_motion=detector_motion,
        eiger=eiger,
        zebra_fast_grid_scan=fast_grid_scan,
        flux=flux,
        oav=oav,
        pin_tip_detection=ophyd_pin_tip_detection,
        smargon=smargon,
        synchrotron=synchrotron,
        s4_slit_gaps=s4_slit_gaps,
        undulator=undulator,
        xbpm_feedback=xbpm_feedback,
        zebra=zebra,
        zocalo=zocalo,
        dcm=dcm,
        robot=MagicMock(spec=BartRobot),
        sample_shutter=zebra_shutter,
    )


@pytest.fixture
async def hyperion_grid_detect_xrc_devices(grid_detect_xrc_devices):
    composite = cast(HyperionGridDetectThenXRayCentreComposite, grid_detect_xrc_devices)
    composite.panda = MagicMock(spec=HDFPanda)
    composite.panda_fast_grid_scan = MagicMock(spec=PandAFastGridScan)
    return composite



from pathlib import Path
from unittest.mock import ANY, MagicMock, patch

import pytest
from dodal.devices.i24.dual_backlight import BacklightPositions

from mx_bluesky.beamlines.i24.serial.parameters.utils import EmptyMapError
from mx_bluesky.beamlines.i24.serial.setup_beamline import Eiger
from mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans import (
    gui_gonio_move_on_click,
    gui_move_backlight,
    gui_move_detector,
    gui_run_chip_collection,
    gui_sleep,
    gui_stage_move_on_click,
)

from ..conftest import fake_generator


@patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.bps.sleep")
def test_gui_sleep(fake_sleep, RE):
    RE(gui_sleep(3))

    assert fake_sleep.call_count == 3


@patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.caput")
@patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.SSX_LOGGER")
async def test_gui_move_detector(mock_logger, fake_caput, detector_stage, RE):
    RE(gui_move_detector("eiger", detector_stage))
    fake_caput.assert_called_once_with("BL24I-MO-IOC-13:GP101", "eiger")

    assert await detector_stage.y.user_readback.get_value() == 59.0
    mock_logger.debug.assert_called_once()


@patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.bps.rd")
@patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.bps.mv")
def test_gui_gonio_move_on_click(fake_mv, fake_rd, RE):
    fake_rd.side_effect = [fake_generator(1.25), fake_generator(1.25)]

    with (
        patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.i24.oav"),
        patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.i24.vgonio"),
    ):
        RE(gui_gonio_move_on_click((10, 20)))

    fake_mv.assert_called_with(ANY, 0.0125, ANY, 0.025)


@patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.get_detector_type")
def test_gui_run_chip_collection_raises_error_for_empty_map(
    mock_det_type,
    RE,
    pmac,
    zebra,
    aperture,
    backlight,
    beamstop,
    detector_stage,
    shutter,
    dcm,
    mirrors,
    eiger_beam_center,
):
    mock_det_type.side_effect = [fake_generator(Eiger())]
    device_list = [
        pmac,
        zebra,
        aperture,
        backlight,
        beamstop,
        detector_stage,
        shutter,
        dcm,
        mirrors,
        eiger_beam_center,
    ]
    with pytest.raises(EmptyMapError):
        RE(
            gui_run_chip_collection(
                "/path/",
                "chip",
                0.01,
                1300,
                0.3,
                1,
                "Oxford",
                "Lite",
                [],
                False,
                "Short1",
                0.01,
                0.005,
                0.0,
                *device_list,
            )
        )


@patch(
    "mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans._move_on_mouse_click_plan"
)
def test_gui_stage_move_on_click(fake_move_plan, oav, pmac, RE):
    RE(gui_stage_move_on_click((200, 200), oav, pmac))
    fake_move_plan.assert_called_once_with(oav, pmac, (200, 200))


@pytest.mark.parametrize("position", ["In", "Out", "White In"])
@patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.SSX_LOGGER")
async def test_gui_move_backlight(mock_logger, position, backlight, RE):
    RE(gui_move_backlight(position, backlight))

    assert (
        await backlight.backlight_position.pos_level.get_value()
        == BacklightPositions(position)
    )
    mock_logger.debug.assert_called_with(f"Backlight moved to {position}")


@patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.DCID")
@patch("mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.get_detector_type")
@patch(
    "mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans._read_visit_directory_from_file"
)
def test_setup_tasks_in_gui_run_chip_collection(
    mock_read_visit,
    mock_det_type,
    mock_dcid,
    RE,
    pmac,
    zebra,
    aperture,
    backlight,
    beamstop,
    detector_stage,
    shutter,
    dcm,
    mirrors,
    eiger_beam_center,
    dummy_params_without_pp,
):
    mock_read_visit.return_value = Path("/tmp/dls/i24/fixed/foo")
    mock_det_type.side_effect = [fake_generator(Eiger())]
    device_list = [
        pmac,
        zebra,
        aperture,
        backlight,
        beamstop,
        detector_stage,
        shutter,
        dcm,
        mirrors,
        eiger_beam_center,
    ]

    expected_params = dummy_params_without_pp
    expected_params.pre_pump_exposure_s = 0.0

    with patch(
        "mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.run_plan_in_wrapper",
        MagicMock(return_value=iter([])),
    ) as patch_wrapped_plan:
        with patch(
            "mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans.upload_chip_map_to_geobrick"
        ) as patch_upload:
            RE(
                gui_run_chip_collection(
                    "bar",
                    "chip",
                    0.01,
                    100,
                    1.0,
                    1,
                    "Oxford",
                    "Lite",
                    [1],
                    False,
                    "NoPP",
                    0.0,
                    0.0,
                    0.0,
                    *device_list,
                )
            )

            patch_upload.assert_called_once_with(pmac, [1])
            mock_dcid.assert_called_once()
            patch_wrapped_plan.assert_called_once_with(
                zebra,
                pmac,
                aperture,
                backlight,
                beamstop,
                detector_stage,
                shutter,
                dcm,
                mirrors,
                eiger_beam_center,
                expected_params,
                mock_dcid(),
            )



from unittest.mock import MagicMock, patch

import bluesky.preprocessors as bpp
import pytest
from bluesky.preprocessors import run_decorator
from bluesky.run_engine import RunEngine

from mx_bluesky.common.external_interaction.alerting import Metadata
from mx_bluesky.common.external_interaction.callbacks.sample_handling.sample_handling_callback import (
    SampleHandlingCallback,
)
from mx_bluesky.common.external_interaction.ispyb.exp_eye_store import BLSampleStatus
from mx_bluesky.common.utils.exceptions import (
    CrystalNotFoundException,
    SampleException,
)

TEST_SAMPLE_ID = 123456
TEST_VISIT = "cm12345-1"
TEST_CONTAINER = 8


@pytest.fixture()
def mock_expeye_cls():
    with patch(
        "mx_bluesky.common.external_interaction.callbacks.sample_handling.sample_handling_callback"
        ".ExpeyeInteraction"
    ) as mock_expeye:
        yield mock_expeye


@run_decorator(
    md={
        "metadata": {
            "sample_id": TEST_SAMPLE_ID,
            "container": TEST_CONTAINER,
            "visit": TEST_VISIT,
        },
        "activate_callbacks": ["SampleHandlingCallback"],
    }
)
def plan_with_general_exception(exception_type: type, msg: str):
    yield from []
    raise exception_type(msg)


def plan_for_sample_id(sample_id):
    def plan_with_exception():
        yield from []
        raise SampleException(f"Test exception for sample_id {sample_id}")

    yield from bpp.run_wrapper(
        plan_with_exception(),
        md={
            "metadata": {"sample_id": sample_id},
            "activate_callbacks": ["SampleHandlingCallback"],
        },
    )


def plan_with_exception_from_inner_plan():
    @run_decorator(
        md={
            "metadata": {"sample_id": TEST_SAMPLE_ID},
        }
    )
    def inner_plan():
        yield from []
        raise SampleException("Exception from inner plan")

    @run_decorator(
        md={
            "metadata": {"sample_id": TEST_SAMPLE_ID},
            "activate_callbacks": ["SampleHandlingCallback"],
        }
    )
    @bpp.set_run_key_decorator("outer_plan")
    def outer_plan():
        yield from inner_plan()

    yield from outer_plan()


def plan_with_rethrown_exception():
    @run_decorator(
        md={
            "metadata": {"sample_id": TEST_SAMPLE_ID},
        }
    )
    def inner_plan():
        yield from []
        raise AssertionError("Exception from inner plan")

    @run_decorator(
        md={
            "metadata": {"sample_id": TEST_SAMPLE_ID},
            "activate_callbacks": ["SampleHandlingCallback"],
        }
    )
    @bpp.set_run_key_decorator("outer_plan")
    def outer_plan():
        try:
            yield from inner_plan()
        except AssertionError as e:
            raise SampleException("Exception from outer plan") from e

    yield from outer_plan()


@run_decorator(
    md={
        "metadata": {"sample_id": TEST_SAMPLE_ID},
        "activate_callbacks": ["SampleHandlingCallback"],
    }
)
def plan_with_normal_completion():
    yield from []


@pytest.mark.parametrize(
    "exception_type, expected_sample_status, message",
    [
        [AssertionError, BLSampleStatus.ERROR_BEAMLINE, "Test failure"],
        [SampleException, BLSampleStatus.ERROR_SAMPLE, "Test failure"],
        [CrystalNotFoundException, BLSampleStatus.ERROR_SAMPLE, "Test failure"],
        [AssertionError, BLSampleStatus.ERROR_BEAMLINE, None],
    ],
)
def test_sample_handling_callback_intercepts_general_exception(
    RE: RunEngine,
    exception_type: type,
    expected_sample_status: BLSampleStatus,
    message: str,
    mock_expeye_cls: MagicMock,
):
    callback = SampleHandlingCallback()
    RE.subscribe(callback)

    with pytest.raises(exception_type):
        RE(plan_with_general_exception(exception_type, message))
    mock_expeye_cls.return_value.update_sample_status.assert_called_once_with(
        TEST_SAMPLE_ID, expected_sample_status
    )


def test_sample_handling_callback_closes_run_normally(
    RE: RunEngine, mock_expeye_cls: MagicMock
):
    callback = SampleHandlingCallback()
    RE.subscribe(callback)
    with (
        patch.object(callback, "_record_exception") as record_exception,
    ):
        RE(plan_with_normal_completion())

    record_exception.assert_not_called()


def test_sample_handling_callback_resets_sample_id(
    mock_expeye_cls: MagicMock, RE: RunEngine
):
    mock_expeye = mock_expeye_cls.return_value
    callback = SampleHandlingCallback()
    RE.subscribe(callback)

    with pytest.raises(SampleException):
        RE(plan_for_sample_id(TEST_SAMPLE_ID))
    mock_expeye.update_sample_status.assert_called_once_with(
        TEST_SAMPLE_ID, BLSampleStatus.ERROR_SAMPLE
    )
    mock_expeye.reset_mock()

    with pytest.raises(SampleException):
        RE(plan_for_sample_id(TEST_SAMPLE_ID + 1))
    mock_expeye.update_sample_status.assert_called_once_with(
        TEST_SAMPLE_ID + 1, BLSampleStatus.ERROR_SAMPLE
    )


def test_sample_handling_callback_triggered_only_by_outermost_plan_when_exception_thrown_in_inner_plan(
    mock_expeye_cls: MagicMock, RE: RunEngine
):
    mock_expeye = mock_expeye_cls.return_value
    callback = SampleHandlingCallback()
    RE.subscribe(callback)

    with pytest.raises(SampleException):
        RE(plan_with_exception_from_inner_plan())
    mock_expeye.update_sample_status.assert_called_once_with(
        TEST_SAMPLE_ID, BLSampleStatus.ERROR_SAMPLE
    )


def test_sample_handling_callback_triggered_only_by_outermost_plan_when_exception_rethrown_from_outermost_plan(
    mock_expeye_cls: MagicMock, RE: RunEngine
):
    mock_expeye = mock_expeye_cls.return_value
    callback = SampleHandlingCallback()
    RE.subscribe(callback)

    with pytest.raises(SampleException):
        RE(plan_with_rethrown_exception())
    mock_expeye.update_sample_status.assert_called_once_with(
        TEST_SAMPLE_ID, BLSampleStatus.ERROR_SAMPLE
    )


@patch.dict("os.environ", {"BEAMLINE": "i03"})
@pytest.mark.parametrize(
    "exception_type, expect_alert, message",
    [
        [AssertionError, True, "Test failure"],
        [SampleException, False, "Test failure"],
        [CrystalNotFoundException, False, "Test failure"],
        [AssertionError, True, None],
    ],
)
def test_sample_handling_callback_raises_an_alert_when_beamline_error_occurs(
    exception_type: type,
    expect_alert: bool,
    message: str,
    mock_expeye_cls: MagicMock,
    mock_alert_service: MagicMock,
    RE: RunEngine,
):
    callback = SampleHandlingCallback()
    RE.subscribe(callback)

    with pytest.raises(exception_type):
        RE(plan_with_general_exception(exception_type, message))

    if expect_alert:
        mock_alert_service.raise_alert.assert_called_once_with(
            "UDC encountered an error on i03",
            f"Hyperion encountered the following beamline error: {message}",
            {
                Metadata.SAMPLE_ID: str(TEST_SAMPLE_ID),
                Metadata.VISIT: TEST_VISIT,
                Metadata.CONTAINER: str(TEST_CONTAINER),
            },
        )
    else:
        mock_alert_service.raise_alert.assert_not_called()



import dataclasses
from unittest.mock import ANY, MagicMock, call, patch

import bluesky.plan_stubs as bps
import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from bluesky.utils import Msg
from dodal.devices.aperturescatterguard import ApertureValue
from dodal.devices.backlight import InOut
from dodal.devices.mx_phase1.beamstop import BeamstopPositions
from dodal.devices.oav.oav_parameters import OAVParameters
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.smargon import CombinedMove
from ophyd_async.testing import get_mock_put

from mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan import (
    BeamlineSpecificFGSFeatures,
    _fire_xray_centre_result_event,
)
from mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan import (
    ConstructBeamlineSpecificFeatures,
    detect_grid_and_do_gridscan,
    grid_detect_then_xray_centre,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    ispyb_activation_wrapper,
)
from mx_bluesky.common.parameters.constants import (
    DocDescriptorNames,
    PlanGroupCheckpointConstants,
)
from mx_bluesky.common.parameters.gridscan import SpecifiedThreeDGridScan
from mx_bluesky.hyperion.parameters.device_composites import (
    GridDetectThenXRayCentreComposite,
)
from mx_bluesky.hyperion.parameters.gridscan import (
    GridScanWithEdgeDetect,
    HyperionSpecifiedThreeDGridScan,
)

from ....conftest import (
    OavGridSnapshotTestEvents,
)
from ...hyperion.experiment_plans.conftest import (
    FLYSCAN_RESULT_LOW,
    FLYSCAN_RESULT_MED,
    sim_fire_event_on_open_run,
)


def _fake_flyscan(*args):
    yield from _fire_xray_centre_result_event([FLYSCAN_RESULT_MED, FLYSCAN_RESULT_LOW])


@pytest.fixture()
def construct_beamline_specific(
    beamline_specific: BeamlineSpecificFGSFeatures,
) -> ConstructBeamlineSpecificFeatures:
    return lambda xrc_composite, xrc_parameters: beamline_specific


@pytest.mark.timeout(2)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.common_flyscan_xray_centre",
    autospec=True,
)
async def test_detect_grid_and_do_gridscan_in_real_RE(
    mock_flyscan: MagicMock,
    pin_tip_detection_with_found_pin: PinTipDetection,
    grid_detect_xrc_devices: GridDetectThenXRayCentreComposite,
    RE: RunEngine,
    test_full_grid_scan_params: GridScanWithEdgeDetect,
    test_config_files: dict,
    construct_beamline_specific: ConstructBeamlineSpecificFeatures,
):
    composite = grid_detect_xrc_devices
    RE(
        ispyb_activation_wrapper(
            _do_detect_grid_and_gridscan_then_wait_for_backlight(
                composite,
                test_config_files,
                test_full_grid_scan_params,
                construct_beamline_specific,
            ),
            test_full_grid_scan_params,
        )
    )

    # Check backlight was moved IN for grid detect then OUT for gridscan
    backlight_mock = get_mock_put(composite.backlight.position)
    backlight_mock.assert_has_calls(
        [call(InOut.IN, wait=True), call(InOut.OUT, wait=True)],
        any_order=False,
    )
    assert backlight_mock.call_count == 2

    # Check aperture was moved out of beam for grid detect
    assert (
        call(ApertureValue.OUT_OF_BEAM, wait=True)
        in get_mock_put(
            composite.aperture_scatterguard.selected_aperture
        ).call_args_list
    )
    # Check aperture was changed to SMALL
    assert (
        await composite.aperture_scatterguard.selected_aperture.get_value()
        == ApertureValue.SMALL
    )

    # Check we called out to underlying fast grid scan plan
    mock_flyscan.assert_called_once_with(ANY, ANY, ANY)


@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.GridDetectionCallback",
    autospec=True,
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.create_parameters_for_flyscan_xray_centre",
    autospec=True,
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.change_aperture_then_move_to_xtal",
    autospec=True,
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.common_flyscan_xray_centre",
    autospec=True,
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.grid_detection_plan",
    autospec=True,
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.setup_beamline_for_OAV",
    autospec=True,
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.XRayCentreEventHandler",
    autospec=True,
)
def test_detect_grid_and_do_gridscan_sets_up_beamline_for_OAV(
    mock_event_handler: MagicMock,
    mock_setup_beamline_for_oav: MagicMock,
    mock_grid_detect: MagicMock,
    mock_flyscan: MagicMock,
    mock_change_aperture_and_move: MagicMock,
    mock_create_params: MagicMock,
    mock_grid_detect_callback: MagicMock,
    grid_detect_xrc_devices: GridDetectThenXRayCentreComposite,
    sim_run_engine: RunEngineSimulator,
    test_full_grid_scan_params: GridScanWithEdgeDetect,
    test_config_files: dict,
    construct_beamline_specific: ConstructBeamlineSpecificFeatures,
):
    mock_event_handler.return_value.xray_centre_results = ["dummy"]
    sim_run_engine.add_handler_for_callback_subscribes()
    sim_run_engine.simulate_plan(
        grid_detect_then_xray_centre(
            grid_detect_xrc_devices,
            test_full_grid_scan_params,
            construct_beamline_specific=construct_beamline_specific,
            oav_config=test_config_files["oav_config_json"],
            xrc_params_type=SpecifiedThreeDGridScan,
        ),
    )

    mock_setup_beamline_for_oav.assert_called_once()


def _do_detect_grid_and_gridscan_then_wait_for_backlight(
    composite,
    test_config_files,
    test_full_grid_scan_params,
    construct_beamline_specific_xrc_features,
):
    yield from detect_grid_and_do_gridscan(
        composite,
        parameters=test_full_grid_scan_params,
        oav_params=OAVParameters("xrayCentring", test_config_files["oav_config_json"]),
        xrc_params_type=HyperionSpecifiedThreeDGridScan,
        construct_beamline_specific=construct_beamline_specific_xrc_features,
    )
    yield from bps.wait(PlanGroupCheckpointConstants.GRID_READY_FOR_DC)


@pytest.mark.timeout(2)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.common_flyscan_xray_centre",
    autospec=True,
)
def test_when_full_grid_scan_run_then_parameters_sent_to_fgs_as_expected(
    mock_flyscan: MagicMock,
    grid_detect_xrc_devices: GridDetectThenXRayCentreComposite,
    RE: RunEngine,
    test_full_grid_scan_params: GridScanWithEdgeDetect,
    test_config_files: dict,
    pin_tip_detection_with_found_pin: PinTipDetection,
    construct_beamline_specific: ConstructBeamlineSpecificFeatures,
):
    oav_params = OAVParameters("xrayCentring", test_config_files["oav_config_json"])

    RE(
        ispyb_activation_wrapper(
            detect_grid_and_do_gridscan(
                grid_detect_xrc_devices,
                parameters=test_full_grid_scan_params,
                oav_params=oav_params,
                xrc_params_type=HyperionSpecifiedThreeDGridScan,
                construct_beamline_specific=construct_beamline_specific,
            ),
            test_full_grid_scan_params,
        )
    )

    params: HyperionSpecifiedThreeDGridScan = mock_flyscan.call_args[0][1]

    assert params.detector_params.num_triggers == 180
    assert params.FGS_params.x_axis.full_steps == 15
    assert params.FGS_params.y_axis.end == pytest.approx(-0.0649, 0.001)

    # Parameters can be serialized
    params.model_dump_json()


@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.grid_detection_plan",
    autospec=True,
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.common_flyscan_xray_centre",
    autospec=True,
)
def test_detect_grid_and_do_gridscan_does_not_activate_ispyb_callback(
    mock_flyscan,
    mock_grid_detection_plan,
    grid_detect_xrc_devices: GridDetectThenXRayCentreComposite,
    sim_run_engine: RunEngineSimulator,
    test_full_grid_scan_params: GridScanWithEdgeDetect,
    test_config_files: dict[str, str],
    construct_beamline_specific: ConstructBeamlineSpecificFeatures,
):
    mock_grid_detection_plan.return_value = iter([Msg("save_oav_grids")])
    sim_run_engine.add_handler_for_callback_subscribes()
    sim_run_engine.add_callback_handler_for_multiple(
        "save_oav_grids",
        [
            [
                (
                    "descriptor",
                    OavGridSnapshotTestEvents.test_descriptor_document_oav_snapshot,  # type: ignore
                ),
                (
                    "event",
                    OavGridSnapshotTestEvents.test_event_document_oav_snapshot_xy,  # type: ignore
                ),
                (
                    "event",
                    OavGridSnapshotTestEvents.test_event_document_oav_snapshot_xz,  # type: ignore
                ),
            ]
        ],
    )

    msgs = sim_run_engine.simulate_plan(
        detect_grid_and_do_gridscan(
            grid_detect_xrc_devices,
            test_full_grid_scan_params,
            OAVParameters("xrayCentring", test_config_files["oav_config_json"]),
            xrc_params_type=HyperionSpecifiedThreeDGridScan,
            construct_beamline_specific=construct_beamline_specific,
        )
    )

    activations = [
        msg
        for msg in msgs
        if msg.command == "open_run"
        and "GridscanISPyBCallback" in msg.kwargs["activate_callbacks"]
    ]
    assert not activations


@pytest.fixture()
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.grid_detection_plan",
    autospec=True,
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.common_flyscan_xray_centre",
    autospec=True,
    side_effect=_fake_flyscan,
)
def msgs_from_simulated_grid_detect_then_xray_centre(
    mock_flyscan,
    mock_grid_detection_plan,
    sim_run_engine: RunEngineSimulator,
    grid_detect_xrc_devices: GridDetectThenXRayCentreComposite,
    test_full_grid_scan_params: GridScanWithEdgeDetect,
    test_config_files: dict[str, str],
    construct_beamline_specific: ConstructBeamlineSpecificFeatures,
):
    mock_grid_detection_plan.return_value = iter(
        [
            Msg("save_oav_grids"),
            Msg(
                "open_run",
                run=DocDescriptorNames.FLYSCAN_RESULTS,
                xray_centre_results=[dataclasses.asdict(FLYSCAN_RESULT_MED)],
            ),
        ]
    )

    sim_run_engine.add_handler_for_callback_subscribes()
    sim_fire_event_on_open_run(sim_run_engine, DocDescriptorNames.FLYSCAN_RESULTS)
    sim_run_engine.add_callback_handler_for_multiple(
        "save_oav_grids",
        [
            [
                (
                    "descriptor",
                    OavGridSnapshotTestEvents.test_descriptor_document_oav_snapshot,  # type: ignore
                ),
                (
                    "event",
                    OavGridSnapshotTestEvents.test_event_document_oav_snapshot_xy,  # type: ignore
                ),
                (
                    "event",
                    OavGridSnapshotTestEvents.test_event_document_oav_snapshot_xz,  # type: ignore
                ),
            ]
        ],
    )
    return sim_run_engine.simulate_plan(
        grid_detect_then_xray_centre(
            grid_detect_xrc_devices,
            test_full_grid_scan_params,
            xrc_params_type=SpecifiedThreeDGridScan,
            construct_beamline_specific=construct_beamline_specific,
            oav_config=test_config_files["oav_config_json"],
        )
    )


def test_grid_detect_then_xray_centre_centres_on_the_first_flyscan_result(
    msgs_from_simulated_grid_detect_then_xray_centre: list[Msg],
):
    assert_message_and_return_remaining(
        msgs_from_simulated_grid_detect_then_xray_centre,
        lambda msg: msg.command == "set"
        and msg.obj.name == "smargon"
        and msg.args[0]
        == CombinedMove(
            x=FLYSCAN_RESULT_MED.centre_of_mass_mm[0],
            y=FLYSCAN_RESULT_MED.centre_of_mass_mm[1],
            z=FLYSCAN_RESULT_MED.centre_of_mass_mm[2],
        ),
    )


def test_grid_detect_then_xray_centre_activates_ispyb_callback(
    msgs_from_simulated_grid_detect_then_xray_centre: list[Msg],
):
    assert_message_and_return_remaining(
        msgs_from_simulated_grid_detect_then_xray_centre,
        lambda msg: msg.command == "open_run"
        and "GridscanISPyBCallback" in msg.kwargs["activate_callbacks"],
    )


def test_detect_grid_and_do_gridscan_waits_for_aperture_to_be_prepared_before_moving_in(
    msgs_from_simulated_grid_detect_then_xray_centre: list[Msg],
):
    msgs = assert_message_and_return_remaining(
        msgs_from_simulated_grid_detect_then_xray_centre,
        lambda msg: msg.command == "prepare"
        and msg.obj.name == "aperture_scatterguard"
        and msg.args[0] == ApertureValue.SMALL,
    )

    aperture_prepare_group = msgs[0].kwargs.get("group")

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait"
        and msg.kwargs["group"] == aperture_prepare_group,
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "aperture_scatterguard-selected_aperture"
        and msg.args[0] == ApertureValue.SMALL,
    )


@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.detect_grid_and_do_gridscan"
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.XRayCentreEventHandler"
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.change_aperture_then_move_to_xtal"
)
def test_grid_detect_then_xray_centre_plan_moves_beamstop_into_place(
    mock_change_aperture_then_move_to_xtal: MagicMock,
    mock_events_handler: MagicMock,
    mock_grid_detect_then_xray_centre: MagicMock,
    sim_run_engine: RunEngineSimulator,
    grid_detect_xrc_devices: GridDetectThenXRayCentreComposite,
    test_full_grid_scan_params: GridScanWithEdgeDetect,
    construct_beamline_specific: ConstructBeamlineSpecificFeatures,
    test_config_files: dict,
):
    flyscan_event_handler = MagicMock()
    flyscan_event_handler.xray_centre_results = "dummy"
    mock_events_handler.return_value = flyscan_event_handler

    mock_grid_detect_then_xray_centre.return_value = iter(
        [Msg("grid_detect_then_xray_centre")]
    )
    msgs = sim_run_engine.simulate_plan(
        grid_detect_then_xray_centre(
            grid_detect_xrc_devices,
            test_full_grid_scan_params,
            SpecifiedThreeDGridScan,
            construct_beamline_specific,
            test_config_files["oav_config_json"],
        )
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        predicate=lambda msg: msg.command == "set"
        and msg.obj.name == "beamstop-selected_pos"
        and msg.args[0] == BeamstopPositions.DATA_COLLECTION,
    )

    msgs = assert_message_and_return_remaining(
        msgs, predicate=lambda msg: msg.command == "grid_detect_then_xray_centre"
    )



import threading
from abc import abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from queue import Empty, Queue
from typing import Any

from blueapi.core import BlueskyContext
from bluesky.callbacks.zmq import Publisher
from bluesky.utils import MsgGenerator

from mx_bluesky.common.external_interaction.callbacks.common.log_uid_tag_callback import (
    LogUidTaggingCallback,
)
from mx_bluesky.common.parameters.components import MxBlueskyParameters
from mx_bluesky.common.parameters.constants import Actions, Status
from mx_bluesky.common.utils.exceptions import WarningException
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.common.utils.tracing import TRACER
from mx_bluesky.hyperion.experiment_plans.experiment_registry import PLAN_REGISTRY
from mx_bluesky.hyperion.parameters.constants import CONST


@dataclass
class Command:
    action: Actions
    devices: Any | None = None
    experiment: Callable[[Any, Any], MsgGenerator] | None = None

    def __str__(self):
        return f"Command({self.action}, {self.parameters}"

    parameters: MxBlueskyParameters | None = None


@dataclass
class StatusAndMessage:
    status: str
    message: str = ""

    def __init__(self, status: Status, message: str = "") -> None:
        self.status = status.value
        self.message = message


@dataclass
class ErrorStatusAndMessage(StatusAndMessage):
    exception_type: str = ""


def make_error_status_and_message(exception: Exception):
    return ErrorStatusAndMessage(
        status=Status.FAILED.value,
        message=repr(exception),
        exception_type=type(exception).__name__,
    )


class BaseRunner:
    @abstractmethod
    def shutdown(self):
        """Performs orderly prompt shutdown.
        Aborts the run engine and terminates the loop waiting for messages."""
        pass

    def __init__(self, context: BlueskyContext):
        self.context: BlueskyContext = context
        self.RE = context.run_engine
        # These references are necessary to maintain liveness of callbacks because RE
        # only keeps a weakref
        self._logging_uid_tag_callback = LogUidTaggingCallback()
        self._publisher = Publisher(f"localhost:{CONST.CALLBACK_0MQ_PROXY_PORTS[0]}")

        self.RE.subscribe(self._logging_uid_tag_callback)
        LOGGER.info("Connecting to external callback ZMQ proxy...")
        self.RE.subscribe(self._publisher)


class GDARunner(BaseRunner):
    """Runner that executes plans submitted by Flask requests from GDA."""

    def __init__(
        self,
        context: BlueskyContext,
    ) -> None:
        super().__init__(context)
        self.current_status: StatusAndMessage = StatusAndMessage(Status.IDLE)
        self._last_run_aborted: bool = False
        self._command_queue: Queue[Command] = Queue()

    def start(
        self,
        experiment: Callable,
        parameters: MxBlueskyParameters,
        plan_name: str | None = None,
    ) -> StatusAndMessage:
        """Start a new bluesky plan
        Args:
            experiment: A bluesky plan
            parameters: The parameters to be submitted
            plan_name: Name of the plan that will be used to resolve the composite factory
                to supply devices for the plan, if any are needed"""
        LOGGER.info(f"Started with parameters: {parameters.model_dump_json(indent=2)}")

        devices: Any = (
            PLAN_REGISTRY[plan_name]["setup"](self.context) if plan_name else None
        )

        if (
            self.current_status.status == Status.BUSY.value
            or self.current_status.status == Status.ABORTING.value
        ):
            return StatusAndMessage(Status.FAILED, "Bluesky already running")
        else:
            self.current_status = StatusAndMessage(Status.BUSY)
            self._command_queue.put(
                Command(
                    action=Actions.START,
                    devices=devices,
                    experiment=experiment,
                    parameters=parameters,
                )
            )
            return StatusAndMessage(Status.SUCCESS)

    def stop(self) -> StatusAndMessage:
        """Stop the currently executing plan."""
        if self.current_status.status == Status.ABORTING.value:
            return StatusAndMessage(Status.FAILED, "Bluesky already stopping")
        else:
            self.current_status = StatusAndMessage(Status.ABORTING)
            stopping_thread = threading.Thread(target=self._stopping_thread)
            stopping_thread.start()
            self._last_run_aborted = True
            return StatusAndMessage(Status.ABORTING)

    def shutdown(self):
        """Stops the run engine and the loop waiting for messages."""
        print("Shutting down: Stopping the run engine gracefully")
        self.stop()
        self._command_queue.put(Command(action=Actions.SHUTDOWN))

    def _stopping_thread(self):
        try:
            # abort() causes the run engine to throw a RequestAbort exception
            # inside the plan, which will propagate through the contingency wrappers.
            # When the plan returns, the run engine will raise RunEngineInterrupted
            self.RE.abort()
            self.current_status = StatusAndMessage(Status.IDLE)
        except Exception as e:
            self.current_status = make_error_status_and_message(e)

    def fetch_next_command(self) -> Command:
        """Fetch the next command from the queue, blocks if queue is empty."""
        return self._command_queue.get()

    def try_fetch_next_command(self) -> Command | None:
        """Fetch the next command from the queue or return None if no command available."""
        try:
            return self._command_queue.get(block=False)
        except Empty:
            return None

    def wait_on_queue(self):
        while True:
            command = self.fetch_next_command()
            if command.action == Actions.SHUTDOWN:
                return
            elif command.action == Actions.START:
                if command.experiment is None:
                    raise ValueError("No experiment provided for START")
                try:
                    with TRACER.start_span("do_run"):
                        self.RE(command.experiment(command.devices, command.parameters))

                    self.current_status = StatusAndMessage(Status.IDLE)

                    self._last_run_aborted = False
                except WarningException as exception:
                    LOGGER.warning("Warning Exception", exc_info=True)
                    self.current_status = make_error_status_and_message(exception)
                except Exception as exception:
                    LOGGER.error("Exception on running plan", exc_info=True)

                    if self._last_run_aborted:
                        # Aborting will cause an exception here that we want to swallow
                        self._last_run_aborted = False
                    else:
                        self.current_status = make_error_status_and_message(exception)



import argparse
from enum import StrEnum

from pydantic.dataclasses import dataclass

from mx_bluesky._version import version


class HyperionMode(StrEnum):
    GDA = "gda"
    UDC = "udc"


@dataclass
class HyperionArgs:
    mode: HyperionMode
    dev_mode: bool = False


def _add_callback_relevant_args(parser: argparse.ArgumentParser) -> None:
    """adds arguments relevant to hyperion-callbacks."""
    parser.add_argument(
        "--dev",
        action="store_true",
        help="Use dev options, such as local graylog instances",
    )


def parse_callback_dev_mode_arg() -> bool:
    """Returns the bool representing the 'dev_mode' argument."""
    parser = argparse.ArgumentParser()
    _add_callback_relevant_args(parser)
    args = parser.parse_args()
    return args.dev


def parse_cli_args() -> HyperionArgs:
    """Parses all arguments relevant to hyperion.
    Returns:
         an HyperionArgs dataclass with the fields: (dev_mode: bool)"""
    parser = argparse.ArgumentParser()
    _add_callback_relevant_args(parser)
    parser.add_argument(
        "--version",
        help="Print hyperion version string",
        action="version",
        version=version,
    )
    parser.add_argument(
        "--mode",
        help="Launch in the specified mode (default is 'gda')",
        default=HyperionMode.GDA,
        type=HyperionMode,
        choices=HyperionMode.__members__.values(),
    )
    args = parser.parse_args()
    return HyperionArgs(dev_mode=args.dev or False, mode=args.mode)



import bluesky.plan_stubs as bps
from dodal.devices.zebra.zebra import (
    ArmDemand,
    EncEnum,
    I03Axes,
    RotationDirection,
    Zebra,
)
from dodal.devices.zebra.zebra_controlled_shutter import (
    ZebraShutter,
    ZebraShutterControl,
)

from mx_bluesky.common.parameters.constants import ZEBRA_STATUS_TIMEOUT
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.phase1_zebra.device_setup_plans.setup_zebra import (
    configure_zebra_and_shutter_for_auto_shutter,
)


def arm_zebra(zebra: Zebra):
    yield from bps.abs_set(zebra.pc.arm, ArmDemand.ARM, wait=True)


def tidy_up_zebra_after_rotation_scan(
    zebra: Zebra,
    zebra_shutter: ZebraShutter,
    group="tidy_up_zebra_after_rotation",
    wait=True,
):
    yield from bps.abs_set(zebra.pc.arm, ArmDemand.DISARM, group=group)
    yield from bps.abs_set(
        zebra_shutter.control_mode, ZebraShutterControl.MANUAL, group=group
    )
    if wait:
        yield from bps.wait(group, timeout=ZEBRA_STATUS_TIMEOUT)


def setup_zebra_for_rotation(
    zebra: Zebra,
    zebra_shutter: ZebraShutter,
    axis: EncEnum = I03Axes.OMEGA,
    start_angle: float = 0,
    scan_width: float = 360,
    shutter_opening_deg: float = 2.5,
    shutter_opening_s: float = 0.04,
    direction: RotationDirection = RotationDirection.POSITIVE,
    group: str = "setup_zebra_for_rotation",
    wait: bool = True,
):
    """Set up the Zebra to collect a rotation dataset. Any plan using this is
    responsible for setting the smargon velocity appropriately so that the desired
    image width is achieved with the exposure time given here.

    Parameters:
        zebra:              The zebra device to use
        axis:               I03 axes enum representing which axis to use for position
                            compare. Currently always omega.
        start_angle:        Position at which the scan should begin, in degrees.
        scan_width:         Total angle through which to collect, in degrees.
        shutter_opening_deg:How many degrees of rotation it takes for the fast shutter
                            to open. Increases the gate width.
        shutter_opening_s:  How many seconds it takes for the fast shutter to open. The
                            detector pulse is delayed after the shutter signal by this
                            amount.
        direction:          RotationDirection enum for positive or negative.
                            Defaults to Positive.
        group:              A name for the group of statuses generated
        wait:               Block until all the settings have completed
    """

    if not isinstance(direction, RotationDirection):
        raise ValueError(
            "Disallowed rotation direction provided to Zebra setup plan. "
            "Use RotationDirection.POSITIVE or RotationDirection.NEGATIVE."
        )
    yield from bps.abs_set(zebra.pc.dir, direction.value, group=group)
    LOGGER.info("ZEBRA SETUP: START")
    # Set gate start, adjust for shutter opening time if necessary
    LOGGER.info(f"ZEBRA SETUP: degrees to adjust for shutter = {shutter_opening_deg}")
    LOGGER.info(f"ZEBRA SETUP: start angle start: {start_angle}")
    LOGGER.info(f"ZEBRA SETUP: start angle adjusted, gate start set to: {start_angle}")
    yield from bps.abs_set(zebra.pc.gate_start, start_angle, group=group)
    # set gate width to total width
    yield from bps.abs_set(
        zebra.pc.gate_width, scan_width + shutter_opening_deg, group=group
    )
    LOGGER.info(
        f"Pulse start set to shutter open time, set to: {abs(shutter_opening_s)}"
    )
    yield from bps.abs_set(zebra.pc.pulse_start, abs(shutter_opening_s), group=group)
    # Set gate position to be angle of interest
    yield from bps.abs_set(zebra.pc.gate_trigger, axis.value, group=group)
    # Set shutter to automatic and to trigger via PC_GATE
    yield from configure_zebra_and_shutter_for_auto_shutter(
        zebra, zebra_shutter, zebra.mapping.sources.PC_GATE, group=group
    )
    # Trigger the detector with a pulse
    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_DETECTOR],
        zebra.mapping.sources.PC_PULSE,
        group=group,
    )
    # Don't use the fluorescence detector
    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_XSPRESS3],
        zebra.mapping.sources.DISCONNECT,
        group=group,
    )
    yield from bps.abs_set(
        zebra.output.pulse_1.input, zebra.mapping.sources.DISCONNECT, group=group
    )
    LOGGER.info(f"ZEBRA SETUP: END - {'' if wait else 'not'} waiting for completion")
    if wait:
        yield from bps.wait(group, timeout=ZEBRA_STATUS_TIMEOUT)


def setup_zebra_for_panda_flyscan(
    zebra: Zebra,
    zebra_shutter: ZebraShutter,
    group="setup_zebra_for_panda_flyscan",
    wait=True,
):
    # Forwards eiger trigger signal from panda
    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_DETECTOR],
        zebra.mapping.sources.IN1_TTL,
        group=group,
    )

    # Set shutter to automatic and to trigger via motion controller GPIO signal (IN4_TTL)
    yield from configure_zebra_and_shutter_for_auto_shutter(
        zebra, zebra_shutter, zebra.mapping.sources.IN4_TTL, group=group
    )

    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_XSPRESS3],
        zebra.mapping.sources.DISCONNECT,
        group=group,
    )

    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_PANDA],
        zebra.mapping.sources.IN3_TTL,
        group=group,
    )  # Tells panda that motion is beginning/changing direction

    if wait:
        yield from bps.wait(group, timeout=ZEBRA_STATUS_TIMEOUT)
